# Copyright (C) 2021-2026 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import datetime
import pandas as pd
import os
import numpy as np
import itertools
import json
import logging
import time
import shutil
import subprocess
import sys
import statistics

from esmecata.utils import is_valid_dir, get_domain_or_superkingdom_from_ncbi_tax_database
from esmecata import __version__ as esmecata_version
from esmecata.core.annotation import extract_protein_cluster, create_dataset_annotation_file
from esmecata.core.clustering import get_proteomes_tax_id_name

from ete4 import NCBITaxa
from ete4 import __version__ as ete4_version

from Bio import __version__ as biopython_version
from Bio import SeqIO

logger = logging.getLogger(__name__)


def get_eggnog_version():
    """Returns version of eggnog-mapper.

    Returns:
        eggnog_version (str): version of eggnog-mapper.
    """
    eggnog_version = None
    response = subprocess.Popen(['emapper.py', '-v'], stdout=subprocess.PIPE, start_new_session=True, universal_newlines="")
    for eggnog_line in response.stdout:
        eggnog_line_decoded = eggnog_line.decode('utf-8')
        if 'emapper-' in eggnog_line_decoded:
            eggnog_version = eggnog_line_decoded.split(' / ')[0].replace('emapper-', '')

    if eggnog_version is None:
        logger.critical('|EsMeCaTa|annotation-eggnog| esmecata could not find the version of eggnog-mapper.')
        logger.critical('|EsMeCaTa|annotation-eggnog| It is possibly an issue with the installation of eggnog-mapper (maybe it is not in the PATH). Or it can be due to a change in the output of emapper.py -v command.')
        sys.exit()

    return eggnog_version


def call_to_emapper(input_path, taxon_name, output_dir, temporary_dir, eggnog_database_path, nb_core, no_dbmem=False, taxon_id_scope=None):
    """ Calls eggnog-mapper on the protein clusters.

    Args:
        input_path (str): pathname to the fasta file containing consensus sequences for protein clusters.
        taxon_name (str): taxon name associated with the protein sequences.
        output_dir (str)): path to the output folder.
        temporayyr_dir (str): path to temporary folder.
        eggnog_database_path (str): pathname to eggnog database folder.
        nb_core (int): number of CPU-cores to use with eggnog-mapper.
        no_dbmem (bool): Boolean to choose to not load eggnog database in memory.
        taxon_id_scope (str): String of taxon IDs associated with query.
    """
    eggnog_starttime = time.time()

    # Use override to avoid issue if already present annotation (if there was a failed run for example).
    eggnog_cmds = ['emapper.py', '--cpu', nb_core, '-i', input_path, '--output', taxon_name,
                '--data_dir', eggnog_database_path, '--itype', 'proteins', '--output_dir', output_dir,
                '--temp_dir', temporary_dir, '--override', '--pident', '40', '--query_cover', '20', '--subject_cover', '20']
    # Use dbmem for faster run, except if the option --no-dbmem has been used.
    if no_dbmem is False:
        eggnog_cmds.append('--dbmem')
    if taxon_id_scope is not None:
        eggnog_cmds.append('--tax_scope')
        eggnog_cmds.append(taxon_id_scope)

    subprocess.call(eggnog_cmds)

    eggnog_endtime = time.time()
    eggnog_duration = eggnog_endtime - eggnog_starttime
    logger.debug('|EsMeCaTa|annotation-eggnog| Emapper complete in {0}s.'.format(eggnog_duration))


def compute_stat_annotation(annotation_reference_folder, stat_file=None):
    """Compute stat associated with the number of proteome for each taxonomic affiliations.

    Args:
        annotation_reference_folder (str): pathname to the annotation reference folder containing annotations for each cluster
        stat_file (str): pathname to the tsv stat file

    Returns:
        annotation_numbers (dict): dict containing observation names (as key) associated with GO Terms and EC (as value)
    """
    annotation_numbers = {}
    annotated_sequenced_numbers = {}
    not_annotated_sequenced_numbers = {}
    taxon_nb_annotation_per_sequences = {}
    sequence_function_associations = {}
    sequence_go_associations = {}
    sequence_ec_associations = {}
    for infile in os.listdir(annotation_reference_folder):
        if '.tsv' in infile:
            observation_name = infile.replace('.tsv','')
            annotation_input_file_path = os.path.join(annotation_reference_folder, infile)
            infile_gos = []
            infile_ecs = []
            annotated_sequences = []
            not_annotated_sequences = []
            nb_annotation_per_sequences = []
            nb_gos_per_sequence = []
            nb_ecs_per_sequence = []
            sequence_function_association = 0
            with open(annotation_input_file_path, 'r') as open_annotation_input_file_path:
                csvreader = csv.DictReader(open_annotation_input_file_path, delimiter='\t')
                for line in csvreader:
                    protein_cluster = line['protein_cluster']
                    gos = [go for go in line['GO'].split(',') if go != '']
                    ecs = [ec for ec in line['EC'].split(',') if ec != '']
                    infile_gos.extend(gos)
                    infile_ecs.extend(ecs)
                    if len(gos) == 0 and len(ecs) == 0:
                        not_annotated_sequences.append(protein_cluster)
                    else:
                        annotated_sequences.append(protein_cluster)
                    nb_annotation_per_sequence = len(gos+ecs)
                    nb_annotation_per_sequences.append(nb_annotation_per_sequence)
                    nb_gos_per_sequence.append(len(gos))
                    nb_ecs_per_sequence.append(len(ecs))
                    sequence_function_association += len(gos+ecs)
            infile_gos = set([go for go in infile_gos])
            infile_ecs = set([ec for ec in infile_ecs])
            annotation_numbers[observation_name] = (len(infile_gos), len(infile_ecs))
            annotated_sequenced_numbers[observation_name] = len(annotated_sequences)
            not_annotated_sequenced_numbers[observation_name] = len(not_annotated_sequences)
            taxon_nb_annotation_per_sequences[observation_name] = statistics.fmean(nb_annotation_per_sequences)
            sequence_function_associations[observation_name] = sequence_function_association
            sequence_go_associations[observation_name] = statistics.fmean(nb_gos_per_sequence)
            sequence_ec_associations[observation_name] = statistics.fmean(nb_ecs_per_sequence)

    if stat_file:
        with open(stat_file, 'w') as stat_file_open:
            csvwriter = csv.writer(stat_file_open, delimiter='\t')
            csvwriter.writerow(['observation_name', 'Number_go_terms', 'Number_ecs', 'Number_sequence_function_association',
                                'Number_annotated_sequences', 'Number_not_annotated_sequences', 'Mean_annotations_per_sequence',
                                'Mean_GOs_per_sequence', 'Mean_ECs_per_sequence'])
            for observation_name in annotation_numbers:
                csvwriter.writerow([observation_name, annotation_numbers[observation_name][0], annotation_numbers[observation_name][1],
                                    sequence_function_associations[observation_name], annotated_sequenced_numbers[observation_name],
                                    not_annotated_sequenced_numbers[observation_name], taxon_nb_annotation_per_sequences[observation_name],
                                    sequence_go_associations[observation_name], sequence_ec_associations[observation_name]])

    return annotation_numbers


def read_annotation(eggnog_outfile:str):
    """Read an eggnog-mapper annotation file and retrieve EC numbers and GO terms by genes.
    Args:
        eggnog_outfile (str): path to eggnog-mapper annotation file

    Returns:
        dict: dict of genes and their annotations as {gene1:{EC:'..,..', GOs:'..,..,'}}
    """
    # Look at the twentieth first rows to find the header.
    with open(eggnog_outfile, 'r') as f:
        twentieth_first_rows = list(itertools.islice(f, 20))
        first_row_after_header = min([index for index, str_row in enumerate(twentieth_first_rows) if not str_row.startswith('#')])
        header_row = first_row_after_header - 1
        headers_row = twentieth_first_rows[header_row].lstrip("#").strip().split('\t')

    # Fix issue when header is incomplete (eggnog before version 2.0).
    if len(headers_row) == 17:
        headers_row.extend(['tax_scope', 'eggNOG_OGs', 'bestOG', 'COG_functional_category', 'eggNOG_free_text'])
    if len(headers_row) == 1:
        headers_row = ['query_name', 'seed_eggNOG_ortholog', 'evalue', 'score', 'eggNOG_OGs', 'max_annot_lvl', 'COG_category', 'Description', 'Preferred_name',
                'GOs', 'EC', 'KEGG_ko', 'KEGG_Pathway', 'KEGG_Module', 'KEGG_Reaction', 'KEGG_rclass', 'BRITE', 'KEGG_TC', 'CAZy', 'BiGG_Reaction', 'PFAMs']

    to_extract_annotations = ['GOs','EC', 'Preferred_name']
    if 'seed_ortholog' in headers_row:
        to_extract_annotations.append('seed_ortholog')
    if 'evalue' in headers_row:
        to_extract_annotations.append('evalue')
    if 'score' in headers_row:
        to_extract_annotations.append('score')
    if 'eggNOG_OGs' in headers_row:
        to_extract_annotations.append('eggNOG_OGs')
    if 'COG_category' in headers_row:
        to_extract_annotations.append('COG_category')
    if 'PFAMs' in headers_row:
        to_extract_annotations.append('PFAMs')
    if 'BiGG_Reaction' in headers_row:
        to_extract_annotations.append('BiGG_Reaction')
    if 'CAZy' in headers_row:
        to_extract_annotations.append('CAZy')
    if 'KEGG_ko' in headers_row:
        to_extract_annotations.append('KEGG_ko')
    if 'KEGG_Pathway' in headers_row:
        to_extract_annotations.append('KEGG_Pathway')
    if 'KEGG_Module' in headers_row:
        to_extract_annotations.append('KEGG_Module')
    if 'KEGG_Reaction' in headers_row:
        to_extract_annotations.append('KEGG_Reaction')

    # Use chunk when reading eggnog file to cope with big file.
    chunksize = 10 ** 6
    for annotation_data in pd.read_csv(eggnog_outfile, sep='\t', comment='#', header=None, dtype=str, chunksize=chunksize):
        annotation_data.replace(np.nan, '', inplace=True)
        # Assign the headers
        annotation_data.columns = headers_row
        if 'query_name' in annotation_data.columns:
            # If a protein is annotated multiple time, selected the match with the highest score.
            annotation_data = annotation_data.sort_values('score', ascending=True).drop_duplicates('query_name').sort_index()
            annotation_dict = annotation_data.set_index('query_name')[to_extract_annotations].to_dict('index')
        # 'query' added for compatibility with eggnog-mapper 2.1.2
        elif 'query' in annotation_data.columns:
            # If a protein is annotated multiple time, selected the match with the highest score.
            annotation_data = annotation_data.sort_values('score', ascending=True).drop_duplicates('query').sort_index()
            annotation_dict = annotation_data.set_index('query')[to_extract_annotations].to_dict('index')
        for key in annotation_dict:
            yield key, annotation_dict[key]


def write_pathologic(base_filename, annotated_proteins, pathologic_output_file, reference_proteins):
    """Write the annotation associated with a cluster after propagation step into pathologic file for run on Pathway Tools.

    Args:
        base_filename (str): observation name
        annotated_proteins (list): list of protein and their annotations as [(protein, {EC:'..,..', GOs:'..,..,'}), (...)]
        reference_proteins (dict): dict containing representative protein IDs (as key) associated with proteins of the cluster
        pathologic_output_file (str): pathname to output pathologic file
    """
    with open(pathologic_output_file, 'w', encoding='utf-8') as element_file:
        element_file.write(';;;;;;;;;;;;;;;;;;;;;;;;;\n')
        element_file.write(';; ' + base_filename + '\n')
        element_file.write(';;;;;;;;;;;;;;;;;;;;;;;;;\n')
        for protein_annots in annotated_proteins:
            if '|' in protein_annots[0]:
                protein = protein_annots[0].split('|')[1]
            else:
                protein = protein_annots[0]

            protein_annot = protein_annots[1]
            element_file.write('ID\t' + protein + '\n')

            protein_name = protein_annot['Preferred_name']
            if protein_name not in ['', '-']:
                element_file.write('NAME\t' + protein_name + '\n')
            else:
                element_file.write('NAME\t' + protein + '\n')
            element_file.write('FUNCTION\tORF\n')
            element_file.write('PRODUCT-TYPE\tP' + '\n')
            element_file.write('PRODUCT-ID\tprot ' + protein + '\n')
            element_file.write('DBLINK\tUNIPROT:' + protein + '\n')
            if 'GOs' in protein_annot:
                gos = [go for go in protein_annot['GOs'].split(',') if go not in ['', '-']]
                for go in gos:
                    element_file.write('GO\t' + go + '\n')
            if 'EC' in protein_annot:
                ecs = [ec for ec in protein_annot['EC'].split(',') if ec not in ['', '-']]
                for ec in ecs:
                    element_file.write('EC\t' + ec + '\n')
            element_file.write('//\n\n')


def write_annotation_reference(protein_annotations, reference_proteins, annotation_reference_file):
    """Write the annotation associated with a cluster after propagation step.

    Args:
        protein_annotations (dict): dict of protein and their annotations as {protein:{EC:'..,..', GOs:'..,..,'}}
        reference_proteins (dict): dict containing representative protein IDs (as key) associated with proteins of the cluster
        annotation_reference_file (str): pathname to output tabulated file
    """
    with open(annotation_reference_file, 'w') as output_tsv:
        csvwriter = csv.writer(output_tsv, delimiter='\t')
        csvwriter.writerow(['protein_cluster', 'cluster_members', 'gene_name', 'GO', 'EC', 'KEGG_reaction'])
        for protein_annot_tuples in protein_annotations:
            if '|' in protein_annot_tuples[0]:
                protein = protein_annot_tuples[0].split('|')[1]
            else:
                protein = protein_annot_tuples[0]
            protein_annot = protein_annot_tuples[1]
            gene_name = protein_annot['Preferred_name']
            cluster_members = ','.join(reference_proteins[protein])
            if 'GOs' in protein_annot:
                gos = [go for go in protein_annot['GOs'].split(',') if go not in ['', '-']]
                gos = ','.join(sorted(gos))
            else:
                gos = ''

            if 'EC' in protein_annot:
                ecs = [ec for ec in protein_annot['EC'].split(',') if ec not in ['', '-']]
                ecs = ','.join(sorted(ecs))
            else:
                ecs = ''

            if 'KEGG_Reaction' in protein_annot:
                keggs = [kegg_rxn for kegg_rxn in protein_annot['KEGG_Reaction'].split(',') if kegg_rxn not in ['', '-']]
                keggs = ','.join(sorted(keggs))
            else:
                keggs = ''

            csvwriter.writerow([protein, cluster_members, gene_name, gos, ecs, keggs])


def merge_fasta_taxa(reference_protein_fasta_path, proteome_tax_id_file, merge_fasta_folder):
    """Merge fasta files contain in reference_proteins_consensus_fasta into bigger fasta files according to their taxonomic rank.
    At this moment, it is merged at the superkingdom level.

    Args:
        reference_protein_fasta_path (str): pathname to the reference_proteins_consensus_fasta folder
        proteome_tax_id_file (str): pathname to the proteomes_tax_id file.
        merge_fasta_folder (str): pathname to the output folder that will be containing the merged fasta

    Returns:
        obs_name_superkingdom (dict): superkingdom as value and the proteomes_tax_id_name associated with them
        taxa_names (dict): list of superkingdom
    """
    if not os.path.exists(reference_protein_fasta_path):
        sys.exit('|EsMeCaTa|annotation-eggnog| Error missing reference_proteins_consensus_fasta in esmecata clustering folder.')
    ncbi = NCBITaxa()
    obs_name_tax_ids = get_proteomes_tax_id_name(proteome_tax_id_file, 'tax_id')
    proteomes_tax_id_names = get_proteomes_tax_id_name(proteome_tax_id_file)

    domain_superkingdom_tax_rank_name = get_domain_or_superkingdom_from_ncbi_tax_database()

    obs_name_superkingdom = {}
    for obs_name in obs_name_tax_ids:
        tax_id = obs_name_tax_ids[obs_name]
        tax_id_lineages = ncbi.get_lineage(tax_id)
        rank_tax_id_lineages = ncbi.get_rank(tax_id_lineages)
        superkingdom = str([tax_id for tax_id in rank_tax_id_lineages if rank_tax_id_lineages[tax_id] == domain_superkingdom_tax_rank_name][0])
        proteomes_tax_id_name = proteomes_tax_id_names[obs_name]
        proteomes_tax_id_name_fasta = os.path.join(reference_protein_fasta_path, proteomes_tax_id_name + '.faa')
        # Check if proteomes_tax_id_name_fasta exists, because if it contains 0 proteins, there will not be a fasta file associated.
        if os.path.exists(proteomes_tax_id_name_fasta):
            if superkingdom not in obs_name_superkingdom:
                obs_name_superkingdom[superkingdom] = [proteomes_tax_id_name]
            else:
                obs_name_superkingdom[superkingdom].append(proteomes_tax_id_name)

    for superkingdom in obs_name_superkingdom:
        obs_name_superkingdom[superkingdom] = set(obs_name_superkingdom[superkingdom])
        merge_fasta_superkingdom_filepath = os.path.join(merge_fasta_folder, superkingdom+'.faa')
        with open(merge_fasta_superkingdom_filepath, 'a') as output_file:
            for proteomes_tax_id_name in obs_name_superkingdom[superkingdom]:
                proteomes_tax_id_name_fasta = os.path.join(reference_protein_fasta_path, proteomes_tax_id_name + '.faa')
                with open(proteomes_tax_id_name_fasta, 'r') as input_file:
                    output_file.write(input_file.read())

    taxa_names = list(obs_name_superkingdom.keys())
    return obs_name_superkingdom, taxa_names


def merged_retrieve_annotation(proteomes_tax_id_names, obs_name_superkingdom, eggnog_output_folder, reference_protein_path,
                               reference_protein_fasta_path, pathologic_folder, annotation_reference_folder):
    """Retrieve annotations from the predictions of eggnog-mapper on the merged fasta files.
    Then write output files.

    Args:
        proteomes_tax_id_names (dict): dict containing observation names (as key) associated with tax name + tax_id used for proteomes (as value)
        obs_name_superkingdom (dict): superkingdom as value and the proteomes_tax_id_name associated with them
        eggnog_output_folder (str): pathname to the output folder of eggnog-mapper
        reference_protein_path (str): pathname to the reference_protein folder of clustering
        reference_protein_fasta_path (str): pathname to the reference_proteins_consensus_fasta folder
        pathologic_folder (str): pathname to the pathologic folder
        annotation_reference_folder (str): pathname to the annotation_reference folder
    """
    for superkingdom in obs_name_superkingdom:
        eggnog_mapper_annotation_file = os.path.join(eggnog_output_folder, superkingdom + '.emapper.annotations')
        annotated_proteins = {prot_id: prot_annot for prot_id, prot_annot in read_annotation(eggnog_mapper_annotation_file)}
        for proteomes_tax_id_name in obs_name_superkingdom[superkingdom]:
            obs_names = [obs_name for obs_name in proteomes_tax_id_names if proteomes_tax_id_names[obs_name] == proteomes_tax_id_name]
            proteomes_tax_id_fasta = os.path.join(reference_protein_fasta_path, proteomes_tax_id_name + '.faa')
            protein_ids = [record.id for record in SeqIO.parse(proteomes_tax_id_fasta, 'fasta')]
            reference_protein_pathname = os.path.join(reference_protein_path, proteomes_tax_id_name+'.tsv')
            reference_proteins, set_proteins = extract_protein_cluster(reference_protein_pathname)

            if len(reference_proteins) > 0:
                sub_annotated_proteins = [(protein_id, annotated_proteins[protein_id]) for protein_id in protein_ids if protein_id in annotated_proteins]

                for obs_name in obs_names:
                    gos = [go for protein_id, protein_annot in sub_annotated_proteins for go in protein_annot['GOs'].split(',') if go not in ['', '-']]
                    unique_gos = set(gos)
                    ecs = [ec for protein_id, protein_annot in sub_annotated_proteins for ec in protein_annot['EC'].split(',') if ec not in ['', '-']]
                    unique_ecs = set(ecs)
                    logger.info('|EsMeCaTa|annotation-eggnog| %d Go Terms (with %d unique GO Terms) and %d EC numbers (with %d unique EC) associated with %s.', len(gos),
                                                                                                            len(unique_gos), len(ecs), len(unique_ecs), obs_name)

                    # Create annotation reference file.
                    annotation_reference_file = os.path.join(annotation_reference_folder, obs_name+'.tsv')
                    if not os.path.exists(annotation_reference_file):
                        write_annotation_reference(sub_annotated_proteins, reference_proteins, annotation_reference_file)

                    # Create pathologic files.
                    pathologic_organism_folder = os.path.join(pathologic_folder, obs_name)
                    # Add _1 to pathologic file as genetic element cannot have the same name as the organism.
                    pathologic_file = os.path.join(pathologic_organism_folder, obs_name+'_1.pf')

                    if not os.path.exists(pathologic_file):
                        if len(sub_annotated_proteins) > 0:
                            is_valid_dir(pathologic_organism_folder)
                            write_pathologic(obs_name, sub_annotated_proteins, pathologic_file, reference_proteins)
                        elif len(sub_annotated_proteins) == 0:
                            logger.critical('|EsMeCaTa|annotation-eggnog| No reference proteins for %s, esmecata will not create a pathologic folder for it.', obs_name)

                    else:
                        logger.critical('|EsMeCaTa|annotation-eggnog| Annotation already performed for %s.', obs_name)
            else:
                logger.critical('|EsMeCaTa|annotation-eggnog| No consensus proteins for %s, annotation will not be performed on it.', proteomes_tax_id_name)


def annotate_with_eggnog(input_folder, output_folder, eggnog_database_path, nb_core, eggnog_tmp_dir=None, no_dbmem=False):
    """Write the annotation associated with a cluster after propagation step into pathologic file for run on Pathway Tools.

    Args:
        input_folder (str): pathname to mmseqs clustering output folder as it is the input of annotation.
        output_folder (str): pathname to the output folder.
        eggnog_database_path (str): pathname to eggnog database folder.
        nb_core (int): number of CPU-cores to be used by eggnog-mapper.
        eggnog_tmp_dir (str): pathname to eggnog-mapper temporary folder.
        no_dbmem (bool): Boolean to choose to not load eggnog database in memory.
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|annotation-eggnog| Begin annotation.')

    is_valid_dir(output_folder)

    eggnog_output_folder = os.path.join(output_folder, 'eggnog_output')
    is_valid_dir(eggnog_output_folder)

    annotation_reference_folder = os.path.join(output_folder, 'annotation_reference')
    is_valid_dir(annotation_reference_folder)

    pathologic_folder = os.path.join(output_folder, 'pathologic')
    is_valid_dir(pathologic_folder)

    reference_protein_fasta_path = os.path.join(input_folder, 'reference_proteins_consensus_fasta')

    proteome_tax_id_file = os.path.join(input_folder, 'proteome_tax_id.tsv')
    annotation_taxon_id_file = os.path.join(output_folder, 'proteome_tax_id.tsv')

    merge_fasta_folder = os.path.join(output_folder, 'merge_fasta')
    is_valid_dir(merge_fasta_folder)
    obs_name_superkingdom, taxa_names = merge_fasta_taxa(reference_protein_fasta_path, proteome_tax_id_file, merge_fasta_folder)

    if os.path.exists(annotation_taxon_id_file):
        if not os.path.samefile(proteome_tax_id_file, annotation_taxon_id_file):
            os.remove(annotation_taxon_id_file)
            shutil.copyfile(proteome_tax_id_file, annotation_taxon_id_file)
    else:
        shutil.copyfile(proteome_tax_id_file, annotation_taxon_id_file)
    proteomes_tax_id_names = get_proteomes_tax_id_name(annotation_taxon_id_file)

    reference_protein_path = os.path.join(input_folder, 'reference_proteins')

    # Convert CPU-core int into str.
    nb_core = str(nb_core)
    # Create a dictionary containing metadata.
    options = {'input_folder': input_folder, 'output_folder': output_folder, 'nb_core': nb_core,
               'eggnog_database_path': eggnog_database_path, 'eggnog_tmp_dir': eggnog_tmp_dir,
               'no_dbmem': no_dbmem}

    options['tool_dependencies'] = {}
    options['tool_dependencies']['python_package'] = {}
    options['tool_dependencies']['python_package']['Python_version'] = sys.version
    options['tool_dependencies']['python_package']['esmecata'] = esmecata_version
    options['tool_dependencies']['python_package']['pandas'] = pd.__version__
    options['tool_dependencies']['python_package']['ete4'] = ete4_version
    options['tool_dependencies']['python_package']['biopython'] = biopython_version
    options['tool_dependencies']['eggnog_mapper'] = get_eggnog_version()

    esmecata_metadata = {}
    date = datetime.datetime.now().strftime('%d-%B-%Y %H:%M:%S')
    esmecata_metadata['access_time'] = date
    esmecata_metadata['tool_options'] = options

    # Run eggnog-mapper on taxon name file.
    for taxon_name in taxa_names:
        # Create temporary folder for eggnog-mapper.
        if eggnog_tmp_dir is None:
            eggnog_temporary_dir = os.path.join(output_folder, 'eggnog_temporary')
        else:
            eggnog_temporary_dir = os.path.join(eggnog_tmp_dir, 'eggnog_temporary')
        is_valid_dir(eggnog_temporary_dir)

        # Launch annotation only if there are protein fasta files to annotate.
        fasta_file_path = os.path.join(merge_fasta_folder, taxon_name+'.faa')

        if os.path.exists(fasta_file_path):
            # Check if the annotation by eggnog-mapper has been performed.
            eggnog_mapper_annotation_file = os.path.join(eggnog_output_folder, taxon_name+'.emapper.annotations')
            if not os.path.exists(eggnog_mapper_annotation_file):
                logger.info('|EsMeCaTa|annotation| Launch eggnog-mapper on %s.', taxon_name)
                call_to_emapper(fasta_file_path, taxon_name, eggnog_output_folder, eggnog_temporary_dir, eggnog_database_path, nb_core, no_dbmem, taxon_name)
            else:
                logger.info('|EsMeCaTa|annotation| Results of eggnog-mapper already present for %s.',  taxon_name)

        # If temporary folder exists, remove it.
        if os.path.exists(eggnog_temporary_dir):
            shutil.rmtree(eggnog_temporary_dir)
    logger.info('|EsMeCaTa|annotation| Runs of eggnog-mapper finished.')

    merged_retrieve_annotation(proteomes_tax_id_names, obs_name_superkingdom, eggnog_output_folder, reference_protein_path,
                                reference_protein_fasta_path, pathologic_folder, annotation_reference_folder)

    # Create mpwt taxon ID file.
    clustering_taxon_id = {}
    with open(proteome_tax_id_file, 'r') as input_taxon_id_file:
        taxon_id_csvreader = csv.DictReader(input_taxon_id_file, delimiter='\t')
        for line in taxon_id_csvreader:
            clustering_taxon_id[line['observation_name']] = line['tax_id']

    pathologic_taxon_id_file = os.path.join(pathologic_folder, 'taxon_id.tsv')
    with open(pathologic_taxon_id_file, 'w') as taxon_id_file:
        taxon_id_csvwriter = csv.writer(taxon_id_file, delimiter='\t')
        taxon_id_csvwriter.writerow(['species', 'taxon_id'])
        for species in clustering_taxon_id:
            taxon_id_csvwriter.writerow([species, clustering_taxon_id[species]])

    function_table_file_path = os.path.join(output_folder, 'function_table.tsv')
    create_dataset_annotation_file(annotation_reference_folder, function_table_file_path, 'all')

    stat_file = os.path.join(output_folder, 'stat_number_annotation.tsv')
    compute_stat_annotation(annotation_reference_folder, stat_file)

    endtime = time.time()
    duration = endtime - starttime
    esmecata_metadata['esmecata_annotation_method'] = 'eggnog-mapper'
    esmecata_metadata['esmecata_annotation_duration'] = duration
    eggnog_metadata_file = os.path.join(output_folder, 'esmecata_metadata_annotation.json')
    if os.path.exists(eggnog_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_annotation' in metadata_file]
        eggnog_metadata_file = os.path.join(output_folder, 'esmecata_metadata_annotation_{0}.json'.format(len(metadata_files)))
        with open(eggnog_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)
    else:
        with open(eggnog_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)

    logger.info('|EsMeCaTa|annotation-eggnog| Annotation complete in {0}s.'.format(duration))



# Copyright (C) 2023-2026  Alice Mataigne and Pauline Hamon-Giraud - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import os
import csv

from typing import Dict, List, Tuple, Any
import plotly.graph_objects as go
import plotly.express as px
import pandas as pd

from esmecata.utils import get_domain_or_superkingdom_from_ncbi_tax_database

# CONSTANTS
# ======================================================================================================================

# TAXONOMIC RANK RELATED CONSTANTS
# --------------------------------


# Due to change in the NCBI Taxonomy database superkingdom has been changed to domain.
# To get the correct one, I search for Bacteria taxonomic ID (2) and comapre the taxonomic rank name to domain or superkingdom.
domain_superkingdom_tax_rank_name = get_domain_or_superkingdom_from_ncbi_tax_database()

RANK2COL = {# KINGDOM (RED)
    domain_superkingdom_tax_rank_name: '#6f1a12', 'kingdom': '#872015', 'subkingdom': '#a2261a',
    # PHYLUM (ORANGE)
    'superphylum': '#b26f15', 'phylum': '#c87c14', 'subphylum': '#de8710', 'infraphylum': '#fa960e',
    # CLASS (YELLOW)
    'superclass': '#c3b21f', 'class': '#d1be14', 'subclass': '#e2cd10', 'infraclass': '#f0db1a',
    # COHORT (KAKHI)
    'cohort': '#a2af21', 'subcohort': '#c1d02e',
    # ORDER (GREEN)
    'superorder': '#699924', 'order': '#7ab12a', 'suborder': '#8ecc33', 'infraorder': '#9dde3e', 'parvorder': '#a9e94c',
    # FAMILY (TURQUOISE)
    'superfamily': '#29b28d', 'family': '#29c198', 'subfamily': '#36d5aa',
    # TRIBE (TURQUOISE - BLUE)
    'tribe': '#3097a4', 'subtribe': '#3eb5c4',
    # GENUS (BLUE)
    'genus': '#2a51ba', 'subgenus': '#3863d7',
    # SECTION (PURPLE)
    'section': '#3a31c6', 'subsection': '#554ce5',
    # SERIES (MAGENTA)
    'series': '#7838c4', 'subseries': '#904ae3',
    # SPECIES (PINK)
    'species group': '#883c91', 'species subgroup': '#9c3ea7', 'species': '#af41bb', 'forma specialis': '#c356cf',
    'subspecies': '#d66be2',
    # (PINK - RED)
    'varietas': '#7e2d48', 'subvariety': '#903855', 'forma': '#a44564', 'serogroup': '#bd5979', 'serotype': '#c96685',
    'strain': '#d57392', 'isolate': '#e085a2',
    # NO RANK (GREY)
    'clade': '#585858', 'environmental samples': '#686868', 'incertae sedis': '#828181', 'unclassified': '#939393',
    'cellular root': '#585858',
    'no rank': '#acacac'}

RANK_SORTED = ['isolate', 'strain', 'serotype', 'serogroup', 'forma', 'subvariety', 'varietas',
               'subspecies', 'forma specialis', 'species', 'species subgroup', 'species group',
               'subseries', 'series',
               'subsection', 'section',
               'subgenus', 'genus',
               'subtribe', 'tribe',
               'subfamily', 'family', 'superfamily',
               'parvorder', 'infraorder', 'suborder', 'order', 'superorder',
               'subcohort', 'cohort',
               'infraclass', 'subclass', 'class', 'superclass',
               'infraphylum', 'subphylum', 'phylum', 'superphylum',
               'subkingdom', 'kingdom', domain_superkingdom_tax_rank_name,
               'clade', 'environmental samples', 'incertae sedis', 'unclassified', 'no rank']

# FIGURE COMPOSITION CONSTANTS
# ----------------------------

DEFAULT_NODE_COLOR = 'rgba(191, 191, 191, 0.3)'
DEFAULT_LINK_COLOR = 'rgba(191, 191, 191, 0.3)'
CHANGE_LINK_COLOR = 'rgba(47, 133, 129, 0.3)'

# DATA CONSTANT
# -------------

# INPUT_DATA KEYS
IN_LST = 'input_lst'
OUT_LST = 'output_lst'
VAL_D = 'val_d'
RANK_D = 'rank_d'

# FIG_DATA KEYS
LABELS = 'labels'
NODE_COLORS = 'node_colors'
RANKS = 'ranks'
SOURCES = 'sources'
TARGETS = 'targets'
VALUES = 'values'
LINK_COLORS = 'link_colors'
TEXT_ANNOT = 'text_annot'
X_POS = 'x_pos'
Y_POS = 'y_pos'

# SUFFIXES
IN_OTU_SUFFIX = ' Organisms'
OUT_TAXA_SUFFIX = ' '


# FUNCTIONS
# ======================================================================================================================
def sort_tax_by_ranks(input_data: Dict[str, Any]) -> Tuple[List[str], List[str]]:
    """ Sort the 'input_lst' and 'output_list' lists of 'input_data' according to taxonomic rank : more specific to
    less specific (~ species --> kingdom)

    Parameters
    ----------
    input_data: Dict[str, Any]
        Dictionary containing input data information

    Returns
    -------
    List[str]
        List of input taxa sorted
    List[str]
        List of output taxa sorted
    """
    input_rvalue = [RANK_SORTED.index(input_data[RANK_D][s]) for s in input_data[IN_LST]]
    output_rvalue = [RANK_SORTED.index(input_data[RANK_D][s]) for s in input_data[OUT_LST]]
    df = pd.DataFrame([input_data[IN_LST], input_data[OUT_LST], input_rvalue, output_rvalue],
                      index=['i_tax', 'o_tax', 'i_val', 'o_val']).T
    df = df.sort_values(by=['i_val', 'o_val', 'o_tax'])
    return list(df['i_tax']), list(df['o_tax'])


def get_input_data(input_file: str) -> Dict[str, Any]:
    """ Read taxonomy_diff.tsv file to extract input and output taxa from a run.

    Parameters
    ----------
    input_file: str
        taxonomy_diff.tsv file from proteome Esmecata step.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing input data information
    """
    input_data = {IN_LST: [], OUT_LST: [], VAL_D: {}, RANK_D: {}}
    with open(input_file, 'r') as f:
        reader = csv.DictReader(f, delimiter='\t')
        for line in reader:
            input_data[IN_LST].append(line['Input Name'])
            input_data[OUT_LST].append(line['Esmecata Name'] + OUT_TAXA_SUFFIX)
            input_data[VAL_D][line['Input Name']] = len(line['Observation Names'].split(';'))
            input_data[RANK_D][line['Input Name']] = line['Input Rank']
            input_data[RANK_D][line['Input Name'] + IN_OTU_SUFFIX] = line['Input Rank']
            input_data[RANK_D][line['Esmecata Name'] + OUT_TAXA_SUFFIX] = line['Esmecata Rank']
    input_lst, output_lst = sort_tax_by_ranks(input_data)
    input_data[IN_LST] = input_lst
    input_data[OUT_LST] = output_lst
    return input_data


def get_fig_parameters(input_data: Dict[str, Any]) -> Dict[str, List[Any]]:
    """ Generate parameters for the sankey figure from input data.

    Parameters
    ----------
    input_data: Dict[str, Any]
        Dictionary containing input data information

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information
    """
    # Init data
    output_set_lst = list(set(input_data[OUT_LST]))
    input_otu = [x + IN_OTU_SUFFIX for x in input_data[IN_LST]]
    output_str = 'Output Diversity'
    input_str = '  '
    labels = input_otu + input_data[IN_LST] + output_set_lst + [str(output_str), str(input_str)]
    nb_otu_input = sum(input_data[VAL_D].values())
    nb_input_div = len(set(input_data[IN_LST]))
    nb_output_div = len(output_set_lst)
    fig_data = {LABELS: labels,
                NODE_COLORS: [RANK2COL[input_data[RANK_D][sp]] if sp in input_data[RANK_D] else DEFAULT_NODE_COLOR
                              for sp in labels],
                RANKS: [input_data[RANK_D][sp] if sp in input_data[RANK_D] else '' for sp in labels],
                SOURCES: [],
                TARGETS: [],
                VALUES: [],
                LINK_COLORS: [],
                TEXT_ANNOT: [f'Input : {nb_otu_input} Organisms',
                             f'Input : {nb_input_div} taxonomic diversity',
                             f'Output : {nb_output_div} taxonomic diversity'],
                X_POS: [0.01] * len(labels),
                Y_POS: [0.0] * len(labels)}
    ii_y = 0.001
    i_y = 0.001
    o_y = 0.001

    # Set Output node position
    fig_data[X_POS][labels.index(output_str)] = 0.99
    fig_data[Y_POS][labels.index(output_str)] = 0.5
    # Make links
    for sp_i in input_data[IN_LST]:
        sp_o = input_data[OUT_LST][input_data[IN_LST].index(sp_i)]
        sp_ii = input_otu[input_data[IN_LST].index(sp_i)]
        # Link Input node (input_str) to Input OTU (sp_ii)
        fig_data, ii_y = link_input_node_to_input_otu(input_data, fig_data, labels, input_str, sp_i, sp_ii,
                                                      nb_otu_input, ii_y)
        # Link Input OTU (sp_ii) to Input Taxa (sp_i)
        fig_data, i_y = link_input_otu_to_input_taxa(fig_data, labels, sp_i, sp_ii, nb_input_div, i_y)
        # Link Input taxa (sp_i) to Output taxa (sp_o) (+ Link Output taxa (sp_o) to Output node (output_str))
        if sp_i != sp_o:
            fig_data, o_y = link_input_taxa_to_output_taxa(input_data, fig_data, labels, sp_i, sp_o, nb_input_div,
                                                           output_str, o_y)
    return fig_data


def set_node_position(node: str, step: float, x_pos: float, increment: float, fig_data: Dict[str, List[Any]],
                      labels: List[str]) -> Tuple[Dict[str, List[Any]], float]:
    """ Set the x and y position of a node in the figure.

    Parameters
    ----------
    node: str
        Label of the node
    step: float
        Proportion of the node height for incrementation of y-axis position
    x_pos: float
        The x-axis position
    increment: float
        Current y-axis position of the iteration
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    labels: List[str]
        List of nodes labels

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information (+ current node position)
    float
        New y-axis position of the iteration
    """
    fig_data[X_POS][labels.index(node)] = x_pos
    fig_data[Y_POS][labels.index(node)] = increment + (step / 2)
    increment += step
    return fig_data, increment


def link_input_node_to_input_otu(input_data: Dict[str, Any], fig_data: Dict[str, List[Any]], labels: List[str],
                                 input_str: str, sp_i: str, sp_ii: str, nb_otu_input: int, ii_y: float)\
        -> Tuple[Dict[str, List[Any]], float]:
    """ Link the input node to an input OTU node (col 1 to col 2).

    Parameters
    ----------
    input_data: Dict[str, Any]
        Dictionary containing input data information
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    labels: List[str]
        List of nodes labels
    input_str: str
        Label of the input node
    sp_i: str
        Label of the input taxa node
    sp_ii: str
        Label of the input OTU node
    nb_otu_input: int
        Total number of OTUs in the input
    ii_y: float
        Previous y-axis position of the input OTU (sp_ii) node

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information (+ current link information)
    float
        New input OTUs y-axis position
    """
    source = labels.index(input_str)
    target = labels.index(sp_ii)
    value = input_data[VAL_D][sp_i]
    fig_data = add_link_data(fig_data, source, target, value, DEFAULT_LINK_COLOR)
    # Set sp_ii position
    step = input_data[VAL_D][sp_i] / nb_otu_input
    return set_node_position(sp_ii, step, 0.2, ii_y, fig_data, labels)


def link_input_otu_to_input_taxa(fig_data: Dict[str, List[Any]], labels: List[str], sp_i: str, sp_ii: str,
                                 nb_input_div: int, i_y: float) -> Tuple[Dict[str, List[Any]], float]:
    """ Link the input OTU nodes do the input taxa nodes (col 2 to col 3).

    Parameters
    ----------
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    labels: List[str]
        List of nodes labels
    sp_i: str
        Label of the input taxa node
    sp_ii: str
        Label of the input OTU node
    nb_input_div: int
        Total number of taxa in the input
    i_y: float
        Previous y-axis position of the input taxa (sp_i) node

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information (+ current link information)
    float
        New input taxa y-axis position
    """
    source = labels.index(sp_ii)
    target = labels.index(sp_i)
    fig_data = add_link_data(fig_data, source, target, 1, DEFAULT_LINK_COLOR)
    # Set sp_i position
    step = 1 / nb_input_div
    return set_node_position(sp_i, step, 0.5, i_y, fig_data, labels)


def link_input_taxa_to_output_taxa(input_data: Dict[str, Any], fig_data: Dict[str, List[Any]], labels: List[str],
                                   sp_i: str, sp_o: str, nb_input_div: int, output_str: str, o_y: float)\
        -> Tuple[Dict[str, List[Any]], float]:
    """ Link the input taxa nodes to output taxa nodes (col 3 to col 4).

    Parameters
    ----------
    input_data: Dict[str, Any]
        Dictionary containing input data information
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    labels: List[str]
        List of nodes labels
    sp_i: str
        Label of the input taxa node
    sp_o: str
        Label of the output taxa node
    nb_input_div: int
        Total number of input taxonomic diversity
    output_str: str
        Label of the output node
    o_y: float
        Previous y-axis position of the output taxa (sp_o) node

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information (+ current link information)
    float
        New output taxa y-axis position

    """
    source = labels.index(sp_i)
    target = labels.index(sp_o)
    value = 1
    if sp_i == sp_o[:-len(OUT_TAXA_SUFFIX)]:
        fig_data = add_link_data(fig_data, source, target, value, DEFAULT_LINK_COLOR)
    else:
        fig_data = add_link_data(fig_data, source, target, value, CHANGE_LINK_COLOR)

    if fig_data[Y_POS][labels.index(sp_o)] == 0.0:
        # Set sp_o position
        step = input_data[OUT_LST].count(sp_o) / nb_input_div
        fig_data, o_y = set_node_position(sp_o, step, 0.8, o_y, fig_data, labels)
        # Link Output taxa (sp_o) to Output node (output_str)
        fig_data = link_output_taxa_to_output_node(fig_data, labels, sp_o, output_str)
    return fig_data, o_y


def link_output_taxa_to_output_node(fig_data: Dict[str, List[Any]], labels: List[str], sp_o: str, output_str: str)\
        -> Dict[str, List[Any]]:
    """ Link the output taxa nodes to the output node (col 4 to col 5).

    Parameters
    ----------
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    labels: List[str]
        List of nodes labels
    sp_o: str
        Label of the output taxa node
    output_str: str
        Label of the output node

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information (+ current link information)
    """
    source = labels.index(sp_o)
    target = labels.index(output_str)
    return add_link_data(fig_data, source, target, 1, DEFAULT_LINK_COLOR)


def add_link_data(fig_data: Dict[str, List[Any]], source: int, target: int, value: int, color: str):
    """ Add a link parameters to fig_data information.

    Parameters
    ----------
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    source: int
        Index of source label
    target: int
        Index of target label
    value: int
        Value of link (flow / size / height)
    color: str
        Color of the link

    Returns
    -------
    Dict[str, List[Any]]
        Dictionary containing figure parameters information (+ current link information)
    """
    fig_data[SOURCES].append(source)
    fig_data[TARGETS].append(target)
    fig_data[VALUES].append(value)
    fig_data[LINK_COLORS].append(color)
    return fig_data


def draw_sankey_fig(fig_data: Dict[str, List[Any]], output: str, show_fig: bool):
    """ Draw the Sankey graph showing compression between input and output taxa of an Esmecata run
    from the fig parameters.

    Parameters
    ----------
    fig_data: Dict[str, List[Any]]
        Dictionary containing figure parameters information
    output: str
        Name of the output file
    show_fig: bool
        True to show figure, False to only save in html
    """
    line_width = 1
    if len(fig_data[LABELS]) > 1000:
        line_width = 0    

    # opacity = 0.4
    # fig_data[""][0]["link"] = [data['data'][0]['node']['color'][src].replace("0.8", str(opacity))
    #                                 for src in data['data'][0]['link']['source']]

    fig = go.Figure(go.Sankey(
        arrangement='snap',
        node={'line': dict(color='black', width=line_width),
              'label': fig_data[LABELS],
              'thickness': 40,
              'pad': 2,
              'color': fig_data[NODE_COLORS],
              'customdata': fig_data[RANKS],
              'hovertemplate': '<b>Name :</b> %{label}<br>'
                               '<b>Rank :</b> %{customdata}<br>'
                               '<extra><b>%{value:,.0f}</b><br></extra>',
              'x': fig_data[X_POS],
              'y': fig_data[Y_POS]},

        link={'source': fig_data[SOURCES],
              'target': fig_data[TARGETS],
              'value': fig_data[VALUES],
              'color': fig_data[LINK_COLORS],
              'line': dict(color='white', width=line_width)}),
    )

    text_pos = [0.2, 0.5, 0.8]
    for i in range(0, 3):
        fig.add_annotation(dict(font=dict(color='darkslategrey', size=18), x=text_pos[i], y=1.05,
                                showarrow=False, text=fig_data[TEXT_ANNOT][i], textangle=0,
                                xanchor='center', xref='paper', yref='paper'))
    # fig.update_layout(height=fig_height)
    # fig.write_html(output)
    if show_fig:
        fig.show()
    return fig


# MAIN FUNCTION
# ======================================================================================================================
def esmecata_compression_taxonomy_file(input_file, output, show_fig):
    """ Draw a Sankey graph showing compression and taxonomic ranks rearrangement between input and output taxa of an
    Esmecata run.

    Parameters
    ----------
    input_file: str
        Path to the taxonomy_diff file in proteomes folder
    output: str
        Name of the output file
    show_fig: bool (default=False)
        True to show figure, False to only save in html
    """
    input_data = get_input_data(input_file)
    fig_data = get_fig_parameters(input_data)
    
    fig_height = 10*len(fig_data[LABELS]) # bug report : Datapane does not uses height correctly if coded inside draw_sankey_fig. Code here instead works (why ?)
    if fig_height < 600:
        fig_height = 600
    
    fig = draw_sankey_fig(fig_data, f'{output}.html', show_fig)
    fig.update_layout(height=fig_height, 
                      font=dict(family="Courier New, monospace",
                                size=14,
                                color="black"))
    fig.write_html(f'{output}.html')

    return fig


def esmecata_compression(run_name: str, output: str, show_fig: bool = False):
    """ Draw a Sankey graph showing compression and taxonomic ranks rearrangement between input and output taxa of an
    Esmecata run.

    Parameters
    ----------
    run_name: str
        Name of the folder of Esmecata results
    output: str
        Name of the output file
    show_fig: bool (default=False)
        True to show figure, False to only save in html
    """
    tax_diff_file = os.path.join(run_name, '0_proteomes', 'taxonomy_diff.tsv')
    fig = esmecata_compression_taxonomy_file(tax_diff_file, output, show_fig)
    return fig


def create_proteomes_barplot(proteomes_df, output_file=None):
    """ Draw a barplot showing the number proteomes according to the taxonomic rank.

    Parameters
    ----------
    df: pandas DataFrame
        Pandas dataframe of the proteomes_tyax_id file.
    output_file: str
        Name of the output file
    """
    df = proteomes_df.drop_duplicates('tax_id_name')
    df['nb_proteomes'] = [len(proteomes.split(',')) for proteomes in df['proteome']]
    tax_rank_in_dataset = df['tax_rank'].tolist()
    order_taxonomic_rank = [taxonomic_rank for taxonomic_rank in RANK_SORTED if taxonomic_rank in tax_rank_in_dataset]
    fig = px.bar(df, x="tax_rank", y="nb_proteomes", category_orders={'tax_rank': order_taxonomic_rank},
                 labels={'x': 'Taxonomic ranks', 'y': 'Number of proteomes'})
    if output_file is not None:
        fig.write_image(output_file)

    return fig



import csv
import os
import shutil
import subprocess
import pandas as pd
import random

from esmecata.core.clustering import make_clustering, filter_protein_cluster, compute_proteome_representativeness_ratio, heap_law_curve_fitting, compute_openness_pan_proteomes

RESULTS = {
    'Cluster_1': {'Number_protein_clusters_kept': 603}
}

def test_filter_protein_cluster_offline():
    output_folder = 'output'
    clustering_input = os.path.join('test_data', 'clustering_input')

    remove_output_folder = False
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
        remove_output_folder = True
        os.mkdir(os.path.join(output_folder, 'computed_threshold'))
        os.mkdir(os.path.join(output_folder, 'reference_proteins'))
        os.mkdir(os.path.join(output_folder, 'cluster_founds'))

    observation_name = 'Cluster_1'
    protein_clusters = {'Q89AE4': ['Q89AE4', 'P57473'],
                        'Q89AY7': ['Q89AY7']}
    observation_name_proteomes = [os.path.join(clustering_input, 'proteomes', 'UP000000601.faa.gz'),
                                os.path.join(clustering_input, 'proteomes', 'UP000001806.faa.gz')]
    clust_threshold = 0.95

    expected_protein = {'Q89AE4'}

    number_proteomes, rep_prot_organims, computed_threshold_cluster = compute_proteome_representativeness_ratio(protein_clusters, observation_name_proteomes)
    protein_cluster_to_keeps = filter_protein_cluster(protein_clusters, number_proteomes, rep_prot_organims, computed_threshold_cluster,
                            clust_threshold)
    
    assert expected_protein == protein_cluster_to_keeps

    clust_threshold = 0
    expected_protein = {'Q89AE4', 'Q89AY7'}
    number_proteomes, rep_prot_organims, computed_threshold_cluster = compute_proteome_representativeness_ratio(protein_clusters, observation_name_proteomes)
    protein_cluster_to_keeps = filter_protein_cluster(protein_clusters, number_proteomes, rep_prot_organims, computed_threshold_cluster,
                            clust_threshold)
    assert expected_protein == protein_cluster_to_keeps

    if remove_output_folder is True:
        shutil.rmtree(output_folder)


def test_make_clustering_offline():
    output_folder = 'clustering_output'
    clustering_input = os.path.join('test_data', 'clustering_input')
    make_clustering(clustering_input, output_folder, nb_core=1, clust_threshold=0.5, mmseqs_options=None, linclust=None, remove_tmp=None)

    expected_results = {}
    output_stat_file = os.path.join(output_folder, 'stat_number_clustering.tsv')
    with open(output_stat_file, 'r') as stat_file_read:
        csvreader = csv.DictReader(stat_file_read, delimiter='\t')
        for line in csvreader:
            expected_results[line['observation_name']] = {}
            expected_results[line['observation_name']]['Number_protein_clusters_kept'] = int(line['Number_protein_clusters_kept'])

    for observation_name in expected_results:
        for data in expected_results[observation_name]:
            assert expected_results[observation_name][data] == RESULTS[observation_name][data]

    shutil.rmtree(output_folder)


def test_clustering_cli_offline():
    output_folder = 'clustering_output'
    clustering_input = os.path.join('test_data', 'clustering_input')
    subprocess.call(['esmecata', 'clustering', '-i', clustering_input, '-o', output_folder, '-c', '1', '-t', '0.5'])
    expected_results = {}
    output_stat_file = os.path.join(output_folder, 'stat_number_clustering.tsv')
    with open(output_stat_file, 'r') as stat_file_read:
        csvreader = csv.DictReader(stat_file_read, delimiter='\t')
        for line in csvreader:
            expected_results[line['observation_name']] = {}
            expected_results[line['observation_name']]['Number_protein_clusters_kept'] = int(line['Number_protein_clusters_kept'])

    for observation_name in expected_results:
        for data in expected_results[observation_name]:
            assert expected_results[observation_name][data] == RESULTS[observation_name][data]

    shutil.rmtree(output_folder)


def test_heap_law_curve_fitting():
    # Test on pangenome from https://doi.org/10.1371/journal.pgen.0030231 by using Table S1
    iteration_nb = 10000
    input_file = os.path.join('test_data', 'pangenome', 'Prochlorococcus_pangenome.tsv')

    # Extract number of genomes and number newly found genes.
    df = pd.read_csv(input_file, sep='\t')
    clustering_gene_genomes = [df[col].dropna().index.tolist() for col in df.columns]
    list_proteome_to_iter = [random.sample(range(len(clustering_gene_genomes)), k=len(clustering_gene_genomes)) for nb_iter in range(iteration_nb)]
    nb_proteomes = []
    nb_new_protein_discovered= []
    for proteome_list in list_proteome_to_iter:
        protein_discovered = set()
        nb_proteome = 0
        for proteome in proteome_list:
            # Get the protein cluters of the new proteome.
            protein_clusters_associated = set(clustering_gene_genomes[proteome])
            nb_proteome += 1
            if protein_discovered == set():
                # If it is the first proteome, all its protein clusters correspond to newly found protein clusters.
                protein_discovered = protein_discovered.union(protein_clusters_associated)
                new_protein_discovered = protein_discovered
            else:
                # Take previously found protein clusters from the other proteomes and extract how many new protein clusters are added by the new proteome.
                new_protein_discovered = protein_clusters_associated - protein_discovered
                protein_discovered = protein_discovered.union(protein_clusters_associated)
            nb_proteomes.append(nb_proteome)
            nb_new_protein_discovered.append(len(new_protein_discovered))

    # Fit curve.
    k, alpha = heap_law_curve_fitting(nb_proteomes, nb_new_protein_discovered)
    # Results should be inferior to 0.9 and close to 0.8 (0.8 found in https://doi.org/10.1016/j.mib.2008.09.006)
    assert alpha < 0.9
    assert alpha > 0.79

def test_compute_openness_pan_proteomes():
    esmecata_computed_threshold_folder = os.path.join('test_data', 'computed_threshold')
    output_openess_file = 'output_openness.tsv'
    compute_openness_pan_proteomes(esmecata_computed_threshold_folder, output_openess_file)
    df = pd.read_csv(output_openess_file, sep='\t')
    predicted_alpha = df['constant_alpha'][0]
    assert 4.1 < predicted_alpha  < 4.6
    os.remove(output_openess_file)

if __name__ == "__main__":
    test_heap_law_curve_fitting()
    test_filter_protein_cluster_offline()
    test_make_clustering_offline()
    test_clustering_cli_offline()
    test_compute_openness_pan_proteomes()



# Copyright (C) 2021-2026 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import argparse
import logging
import os
import sys
import time

from esmecata.core.proteomes import check_proteomes, retrieve_proteomes
from esmecata.core.clustering import make_clustering
from esmecata.core.annotation import annotate_proteins
from esmecata.core.workflow import perform_workflow, perform_workflow_eggnog
from esmecata.core.eggnog import annotate_with_eggnog
from esmecata.core.precomputed import precomputed_parse_affiliation
from esmecata.utils import limited_integer_type, range_limited_float_type, is_valid_dir
from esmecata import __version__ as VERSION

MESSAGE = '''
From taxonomic affiliation to metabolism using Uniprot.
'''
REQUIRES = '''
Steps proteomes and annotation by UniProt requires an internet connection (for REST and SPARQL queries, except if you have a local Uniprot SPARQL endpoint).
Step clustering requires mmseqs2.
Annotation can be performed with UniProt or eggnog-mapper (which is then a requirement if the option is selected).
Precomputed requires the esmecata_database.zip file.
'''

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

def main():
    start_time = time.time()

    parser = argparse.ArgumentParser(
        'esmecata',
        description=MESSAGE + ' For specific help on each subcommand use: esmecata {cmd} --help',
        epilog=REQUIRES
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + VERSION + '\n')

    parent_parser_i_taxon = argparse.ArgumentParser(add_help=False)
    parent_parser_i_taxon.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='Input taxon file (excel, tsv or csv) containing a column associating ID to a taxonomic affiliation (separated by ;).',
        metavar='INPUT_FILE')

    parent_parser_i_clustering_folder = argparse.ArgumentParser(add_help=False)
    parent_parser_i_clustering_folder.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='This input folder of clustering is the output folder of proteomes command.',
        metavar='INPUT_DIR')

    parent_parser_i_annotation_folder = argparse.ArgumentParser(add_help=False)
    parent_parser_i_annotation_folder.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='This input folder of annotation is the output folder of clustering command.',
        metavar='INPUT_DIR')

    parent_parser_i_analysis_folder = argparse.ArgumentParser(add_help=False)
    parent_parser_i_analysis_folder.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='This input folder of analysis is the output folder of annotation command.',
        metavar='INPUT_DIR')

    parent_parser_o = argparse.ArgumentParser(add_help=False)
    parent_parser_o.add_argument(
        '-o',
        '--output',
        dest='output',
        required=True,
        help='Output directory path.',
        metavar='OUPUT_DIR')

    parent_parser_b = argparse.ArgumentParser(add_help=False)
    parent_parser_b.add_argument(
        '-b',
        '--busco',
        dest='busco',
        required=False,
        type=range_limited_float_type,
        help='BUSCO percentage between 0 and 1. This will remove all the proteomes without BUSCO score and the score before the selected ratio of completion.',
        metavar='BUSCO',
        default=0.8)
    parent_parser_taxadb = argparse.ArgumentParser(add_help=False)
    parent_parser_taxadb.add_argument(
        '--ignore-taxadb-update',
        dest='ignore_taxadb_update',
        help='If you have a not up-to-date version of the NCBI taxonomy database with ete4, use this option to bypass the warning message and use the old version.',
        required=False,
        action='store_true',
        default=None)
    parent_parser_all_proteomes = argparse.ArgumentParser(add_help=False)
    parent_parser_all_proteomes.add_argument(
        '--all-proteomes',
        dest='all_proteomes',
        help='Download all proteomes associated with a taxon even if they are no reference proteomes.',
        required=False,
        action='store_true',
        default=None)
    parent_parser_limit_maximal_number_proteomes = argparse.ArgumentParser(add_help=False)
    parent_parser_limit_maximal_number_proteomes.add_argument(
        '-l',
        '--limit-proteomes',
        dest='limit_maximal_number_proteomes',
        required=False,
        type=limited_integer_type,
        help='Choose the maximal number of proteomes after which the tool will select a subset of proteomes instead of using all the available proteomes (default is 99).',
        default=99)
    parent_parser_update_affiliation = argparse.ArgumentParser(add_help=False)
    parent_parser_update_affiliation.add_argument(
        '--update-affiliations',
        dest='update_affiliations',
        help='''If the taxonomic affiliations were assigned from an outdated taxonomic database, this can lead to taxon not be found in ete4 database. \
            This option tries to update the taxonomic affiliations using the lowest taxon name.''',
        required=False,
        action='store_true',
        default=None)
    parent_parser_c = argparse.ArgumentParser(add_help=False)
    parent_parser_c.add_argument(
        '-c',
        '--core',
        dest='core',
        help='Number of CPU-cores for multiprocessing.',
        required=False,
        type=int,
        default=1)
    parent_parser_thr = argparse.ArgumentParser(add_help=False)
    parent_parser_thr.add_argument(
        '-t',
        '--threshold',
        dest='threshold_clustering',
        help='Proportion [0 to 1] of proteomes required to occur in a proteins cluster for that cluster to be kept in core proteome assembly. Default is 0.5.',
        required=False,
        type=range_limited_float_type,
        default=0.5)
    parent_parser_mmseqs_options = argparse.ArgumentParser(add_help=False)
    parent_parser_mmseqs_options.add_argument(
        '-m',
        '--mmseqs',
        dest='mmseqs_options',
        help='String containing mmseqs options for cluster command (except --threads which is already set by --core command and -v). If nothing is given, esmecata will used the option "--min-seq-id 0.3 -c 0.8"',
        required=False,
        type=str,
        default=None)
    parent_parser_linclust = argparse.ArgumentParser(add_help=False)
    parent_parser_linclust.add_argument(
        '--linclust',
        dest='linclust',
        help='Use mmseqs linclust (clustering in linear time) to cluster proteins sequences. It is faster than mmseqs cluster (default behaviour) but less sensitive.',
        required=False,
        action='store_true',
        default=None)
    parent_parser_remove_tmp = argparse.ArgumentParser(add_help=False)
    parent_parser_remove_tmp.add_argument(
        '--remove-tmp',
        dest='remove_tmp',
        help='Delete tmp files to limit the disk space used: files created by mmseqs (in mmseqs_tmp).',
        required=False,
        action='store_true',
        default=None)
    parent_parser_propagate = argparse.ArgumentParser(add_help=False)
    parent_parser_propagate.add_argument(
        '-p',
        '--propagate',
        dest='propagate_annotation',
        help='Proportion [0 to 1] of the occurrence of an annotation to be propagated from the protein of a cluster to the reference protein of the cluster. 0 mean the annotations from all proteins are propagated to the reference and 1 only the annotation occurring in all the proteins of the cluster (default).',
        required=False,
        type=range_limited_float_type,
        default=1)
    parent_parser_uniref = argparse.ArgumentParser(add_help=False)
    parent_parser_uniref.add_argument(
        '--uniref',
        dest='uniref',
        help='Use uniref cluster to extract more annotations from the representative member of the cluster associated with the proteins. Needs the --sparql option.',
        required=False,
        action='store_true',
        default=None)
    parent_parser_expression = argparse.ArgumentParser(add_help=False)
    parent_parser_expression.add_argument(
        '--expression',
        dest='expression',
        help='Extract expression information associated with the proteins. Needs the --sparql option.',
        required=False,
        action='store_true',
        default=None)
    parent_parser_sparql = argparse.ArgumentParser(add_help=False)
    parent_parser_sparql.add_argument(
        '-s',
        '--sparql',
        dest='sparql',
        help='Use sparql endpoint instead of REST queries on Uniprot (to use UniProt PARQL Endpoint, give "uniprot" to this parameter).',
        required=False)
    parent_parser_rank_limit = argparse.ArgumentParser(add_help=False)
    parent_parser_rank_limit.add_argument(
        '-r',
        '--rank-limit',
        dest='rank_limit',
        required=False,
        help='''This option limits the rank used when searching for proteomes. All the ranks superior to the given rank will be ignored. \
            For example, if 'family' is given, only taxon ranks inferior or equal to family will be kept. \
            Look at the readme for more information (and a list of rank names).''',
        default=None)
    parent_parser_minimal_number_proteomes = argparse.ArgumentParser(add_help=False)
    parent_parser_minimal_number_proteomes.add_argument(
        '--minimal-nb-proteomes',
        dest='minimal_number_proteomes',
        required=False,
        type=limited_integer_type,
        help='Choose the minimal number of proteomes to be selected by EsMeCaTa. If a taxon has less proteomes, it will be ignored and a higher taxonomic rank will be used. Default is 5.',
        default=5)
    parent_parser_annotation_file = argparse.ArgumentParser(add_help=False)
    parent_parser_annotation_file.add_argument(
        '--annotation-files',
        dest='annotation_files',
        required=False,
        help='Use UniProt annotation files (uniprot_trembl.txt and uniprot_sprot.txt) to avoid querying UniProt REST API. Need both paths to these files separated by a ",".',
        default=None)
    parent_parser_bioservices = argparse.ArgumentParser(add_help=False)
    parent_parser_bioservices.add_argument(
        '--bioservices',
        dest='option_bioservices',
        help='Use bioservices instead of esmecata functions for protein annotation.',
        required=False,
        action='store_true',
        default=None)
    parent_parser_eggnog_database = argparse.ArgumentParser(add_help=False)
    parent_parser_eggnog_database.add_argument(
        '-e',
        '--eggnog',
        dest='eggnog_database',
        help='Path to eggnog database.',
        required=True)
    parent_parser_taxon_rank = argparse.ArgumentParser(add_help=False)
    parent_parser_taxon_rank.add_argument(
        '--taxon-rank',
        dest='taxon_rank',
        required=False,
        help='Taxon rank to merge organisms (default family).',
        default='family')
    parent_parser_eggnog_tmp_dir = argparse.ArgumentParser(add_help=False)
    parent_parser_eggnog_tmp_dir.add_argument(
        '--eggnog-tmp',
        dest='eggnog_tmp_dir',
        help='Path to eggnog tmp dir.',
        required=False,
        default=None)

    parent_parser_d = argparse.ArgumentParser(add_help=False)
    parent_parser_d.add_argument(
        '-d',
        '--database',
        dest='database',
        required=True,
        nargs="+",
        help='EsMeCaTa precomputed database file path. Multiple precomputed databases can be given, separated by a " ", for example -d "esmecata_db1.zip esmecata_db2.zip".',
        metavar='INPUT_FILE')

    parent_parser_no_dbmem = argparse.ArgumentParser(add_help=False)
    parent_parser_no_dbmem.add_argument(
        '--no-dbmem',
        dest='no_dbmem',
        help='Do not load eggnog database into memory.',
        required=False,
        action='store_true',
        default=False)

    # subparsers
    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands:',
        dest='cmd')

    check_parser = subparsers.add_parser(
        'check',
        help='Check proteomes associated with taxon in Uniprot Proteomes database.',
        parents=[
            parent_parser_i_taxon, parent_parser_o, parent_parser_b,
            parent_parser_taxadb, parent_parser_all_proteomes, parent_parser_sparql,
            parent_parser_limit_maximal_number_proteomes, parent_parser_rank_limit,
            parent_parser_minimal_number_proteomes, parent_parser_update_affiliation,
            parent_parser_bioservices
            ],
        allow_abbrev=False)
    proteomes_parser = subparsers.add_parser(
        'proteomes',
        help='Download proteomes associated with taxon from Uniprot Proteomes.',
        parents=[
            parent_parser_i_taxon, parent_parser_o, parent_parser_b,
            parent_parser_taxadb, parent_parser_all_proteomes, parent_parser_sparql,
            parent_parser_limit_maximal_number_proteomes, parent_parser_rank_limit,
            parent_parser_minimal_number_proteomes, parent_parser_update_affiliation,
            parent_parser_bioservices
            ],
        allow_abbrev=False)
    clustering_parser = subparsers.add_parser(
        'clustering',
        help='Cluster the proteins of the different proteomes of a taxon into a single set of representative shared proteins.',
        parents=[
            parent_parser_i_clustering_folder, parent_parser_o, parent_parser_c,
            parent_parser_thr, parent_parser_mmseqs_options, parent_parser_linclust,
            parent_parser_remove_tmp
            ],
        allow_abbrev=False)
    annotation_uniprot_parser = subparsers.add_parser(
        'annotation_uniprot',
        help='Retrieve protein annotations from Uniprot.',
        parents=[
            parent_parser_i_annotation_folder, parent_parser_o, parent_parser_sparql,
            parent_parser_propagate, parent_parser_uniref, parent_parser_expression,
            parent_parser_annotation_file, parent_parser_bioservices
            ],
        allow_abbrev=False)
    annotation_eggnog_parser = subparsers.add_parser(
        'annotation',
        help='Annotate protein clusters using eggnog-mapper.',
        parents=[
            parent_parser_i_annotation_folder, parent_parser_o, parent_parser_eggnog_database,
            parent_parser_c, parent_parser_eggnog_tmp_dir, parent_parser_no_dbmem
            ],
        allow_abbrev=False)
    workflow_uniprot_parser = subparsers.add_parser(
        'workflow_uniprot',
        help='Run all esmecata steps (proteomes, clustering and annotation).',
        parents=[
            parent_parser_i_taxon, parent_parser_o, parent_parser_b, parent_parser_c,
            parent_parser_taxadb, parent_parser_all_proteomes, parent_parser_sparql,
            parent_parser_remove_tmp, parent_parser_limit_maximal_number_proteomes,
            parent_parser_thr, parent_parser_mmseqs_options, parent_parser_linclust,
            parent_parser_propagate, parent_parser_uniref, parent_parser_expression,
            parent_parser_rank_limit, parent_parser_minimal_number_proteomes,
            parent_parser_annotation_file, parent_parser_update_affiliation,
            parent_parser_bioservices
            ],
        allow_abbrev=False)
    workflow_eggnog_parser = subparsers.add_parser(
        'workflow',
        help='Run all esmecata steps (proteomes, clustering and annotation with eggnog-mapper).',
        parents=[
            parent_parser_i_taxon, parent_parser_o, parent_parser_eggnog_database,
            parent_parser_b, parent_parser_c, parent_parser_taxadb,
            parent_parser_all_proteomes, parent_parser_sparql, parent_parser_remove_tmp,
            parent_parser_limit_maximal_number_proteomes, parent_parser_thr, parent_parser_mmseqs_options,
            parent_parser_linclust, parent_parser_rank_limit, parent_parser_minimal_number_proteomes,
            parent_parser_update_affiliation, parent_parser_bioservices, parent_parser_eggnog_tmp_dir,
            parent_parser_no_dbmem
            ],
        allow_abbrev=False)

    precomputed_parser = subparsers.add_parser(
        'precomputed',
        help='Use precomputed database to create estimated data for the run.',
        parents=[
            parent_parser_i_taxon, parent_parser_d, parent_parser_o,
            parent_parser_rank_limit, parent_parser_update_affiliation,
            parent_parser_thr
            ],
        allow_abbrev=False)

    args = parser.parse_args()

    # If no argument print the help.
    if len(sys.argv) == 1 or len(sys.argv) == 0:
        parser.print_help()
        sys.exit(1)

    is_valid_dir(args.output)

    # add logger in file
    formatter = logging.Formatter('%(message)s')
    log_file_path = os.path.join(args.output, f'esmecata_{args.cmd}.log')
    file_handler = logging.FileHandler(log_file_path, 'w+')
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # set up the default console logger
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if args.cmd in ['proteomes', 'annotation_uniprot', 'workflow_uniprot', 'workflow', 'check']:
        if args.sparql is None:
            uniprot_sparql_endpoint = None
        elif args.sparql == 'uniprot':
            uniprot_sparql_endpoint = 'https://sparql.uniprot.org/sparql'
        else:
            uniprot_sparql_endpoint = args.sparql

    if args.cmd in ['proteomes', 'workflow_uniprot', 'workflow', 'check']:
        if args.busco is not None:
            busco_score = 100*args.busco

    if args.cmd == 'proteomes':
        retrieve_proteomes(args.input, args.output, busco_score, args.ignore_taxadb_update,
                            args.all_proteomes, uniprot_sparql_endpoint, args.limit_maximal_number_proteomes,
                            args.rank_limit, args.minimal_number_proteomes, args.update_affiliations,
                            args.option_bioservices)
    if args.cmd == 'check':
        check_proteomes(args.input, args.output, busco_score, args.ignore_taxadb_update,
                            args.all_proteomes, uniprot_sparql_endpoint, args.limit_maximal_number_proteomes,
                            args.rank_limit, args.minimal_number_proteomes, args.update_affiliations,
                            args.option_bioservices)
    elif args.cmd == 'clustering':
        make_clustering(args.input, args.output, args.core, args.threshold_clustering, args.mmseqs_options,
                        args.linclust, args.remove_tmp)
    elif args.cmd == 'annotation_uniprot':
        annotate_proteins(args.input, args.output, uniprot_sparql_endpoint,
                        args.propagate_annotation, args.uniref, args.expression,
                        args.annotation_files, args.option_bioservices)
    elif args.cmd == 'workflow_uniprot':
        perform_workflow(args.input, args.output, busco_score, args.ignore_taxadb_update,
                            args.all_proteomes, uniprot_sparql_endpoint, args.remove_tmp,
                            args.limit_maximal_number_proteomes, args.rank_limit,
                            args.core, args.threshold_clustering, args.mmseqs_options,
                            args.linclust, args.propagate_annotation, args.uniref,
                            args.expression, args.minimal_number_proteomes, args.annotation_files,
                            args.update_affiliations, args.option_bioservices)
    elif args.cmd == 'annotation':
        annotate_with_eggnog(args.input, args.output, args.eggnog_database, args.core,
                             args.eggnog_tmp_dir, args.no_dbmem)
    elif args.cmd == 'workflow':
        perform_workflow_eggnog(args.input, args.output, args.eggnog_database, busco_score,
                                args.ignore_taxadb_update, args.all_proteomes, uniprot_sparql_endpoint,
                                args.remove_tmp, args.limit_maximal_number_proteomes, args.rank_limit,
                                args.core, args.threshold_clustering, args.mmseqs_options,
                                args.linclust, args.minimal_number_proteomes, args.update_affiliations,
                                args.option_bioservices, args.eggnog_tmp_dir, args.no_dbmem)
    elif args.cmd == 'precomputed':
        precomputed_db = ' '.join(args.database)
        precomputed_parse_affiliation(args.input, precomputed_db, args.output, args.rank_limit, args.update_affiliations,
                                      args.threshold_clustering)

    logger.info("--- Total runtime %.2f seconds ---" % (time.time() - start_time))
    logger.warning(f'--- Logs written in {log_file_path} ---')


if __name__ == '__main__':
    main()



# Copyright (C) 2021-2026 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import argparse
import logging
import os
import sys
import time

from esmecata.gseapy.gseapy_orsum import taxon_rank_annotation_enrichment
from esmecata.utils import is_valid_dir
from esmecata import __version__ as VERSION

MESSAGE = '''
Create enrichment analysis files from esmecata results.
'''
REQUIRES = '''
Requires: bioservices, pronto, gseapy and orsum
'''

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

def main():
    start_time = time.time()

    parser = argparse.ArgumentParser(
        'esmecata_gseapy',
        description=MESSAGE + ' For specific help on each subcommand use: esmecata {cmd} --help',
        epilog=REQUIRES
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + VERSION + '\n')

    parent_parser_f = argparse.ArgumentParser(add_help=False)
    parent_parser_f.add_argument(
        '-f',
        '--folder',
        dest='input_file_folder',
        required=True,
        help='Annotation input folder or file. For folder, it corresponds to EsMeCaTa annotation output folder. For file, it ocrresponds to a function table containing annotation as column and organism as row.',
        metavar='INPUT_FOLDER')

    parent_parser_o = argparse.ArgumentParser(add_help=False)
    parent_parser_o.add_argument(
        '-o',
        '--output',
        dest='output',
        required=True,
        help='Output directory path.',
        metavar='OUPUT_DIR')

    parent_parser_grouping = argparse.ArgumentParser(add_help=False)
    parent_parser_grouping.add_argument(
        '--grouping',
        dest='grouping',
        required=True,
        help='Grouping factor used for enrichment analysis (either "tax_rank", "selected" or "selected_function").',
        metavar='STR',
        default='tax_rank')

    parent_parser_t = argparse.ArgumentParser(add_help=False)
    parent_parser_t.add_argument(
        '--taxon-rank',
        dest='taxon_rank',
        required=False,
        help='Taxon rank to cluster observation names of EsMeCaTa together (default "phylum").',
        metavar='INPUT_TAXON',
        default='phylum')

    parent_parser_taxa_list = argparse.ArgumentParser(add_help=False)
    parent_parser_taxa_list.add_argument(
        '--taxa-list',
        dest='taxa_list',
        required=False,
        help='When using value "selected" for option "grouping", you have to give this file indicating the different groups of taxon to compare.',
        metavar='STR',
        default='tax_rank')

    parent_parser_function_list = argparse.ArgumentParser(add_help=False)
    parent_parser_function_list.add_argument(
        '--function-list',
        dest='function_list',
        required=False,
        help='When using value "selected_function" for option "grouping", you have to give this file indicating the different groups of functions to compare.',
        metavar='STR',
        default='function_list')

    parent_parser_annot_names = argparse.ArgumentParser(add_help=False)
    parent_parser_annot_names.add_argument(
        '--annot-names',
        dest='annot_names_file',
        required=False,
        help='Pathname to json file indicating annotation names.',
        metavar='INPUT_FILE',
        default='download')

    parent_parser_taxon_id = argparse.ArgumentParser(add_help=False)
    parent_parser_taxon_id.add_argument(
        '--taxon-id',
        dest='taxon_id',
        required=False,
        help='Pathname to taxon ID file indicating taxonomic ID associated with organisms name. Required for tax_rank analysis if the annotation input is a file and not an esmecata annotation folder.',
        metavar='INPUT_FILE',
        default=None)

    parent_parser_ko = argparse.ArgumentParser(add_help=False)
    parent_parser_ko.add_argument(
        '--ko',
        dest='ko',
        help='If KEGG Orthologs are in the function table, add this parameter to get their names (require bioservices).',
        required=False,
        action='store_true',
        default=False)

    parent_parser_orsumMinTermSize = argparse.ArgumentParser(add_help=False)
    parent_parser_orsumMinTermSize.add_argument(
        '--orsumMinTermSize',
        dest='orsumMinTermSize',
        required=False,
        help='MinTermSize of orsum.',
        metavar='INT',
        default=None)

    parent_parser_gseapyCutOff = argparse.ArgumentParser(add_help=False)
    parent_parser_gseapyCutOff.add_argument(
        '--gseapyCutOff',
        dest='gseapyCutOff',
        required=False,
        type=float,
        help='Adjust-Pval cutoff for gseapy enrichr, default: 0.05 (--cut-off argument of gseapy).',
        metavar='FLOAT',
        default=0.05)

    # subparsers
    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands:',
        dest='cmd')

    gseapy_taxon_parser = subparsers.add_parser(
        'gseapy_enrichr',
        help='Extract enriched functions from groups (either chosen from tax_rank or manually selected) using gseapy enrichr and orsum.',
        parents=[
            parent_parser_f, parent_parser_o, parent_parser_grouping, parent_parser_t, parent_parser_taxa_list, parent_parser_function_list,
            parent_parser_annot_names, parent_parser_orsumMinTermSize, parent_parser_gseapyCutOff, parent_parser_taxon_id
            ],
        allow_abbrev=False)

    args = parser.parse_args()

    # If no argument print the help.
    if len(sys.argv) == 1 or len(sys.argv) == 0:
        parser.print_help()
        sys.exit(1)

    is_valid_dir(args.output)

    # add logger in file
    formatter = logging.Formatter('%(message)s')
    log_file_path = os.path.join(args.output, f'esmecata_gseapy_{args.cmd}.log')
    file_handler = logging.FileHandler(log_file_path, 'w+')
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # set up the default console logger
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if args.cmd == 'gseapy_enrichr':
        taxon_rank_annotation_enrichment(args.input_file_folder, args.output, args.grouping, taxon_rank=args.taxon_rank, taxa_lists_file=args.taxa_list,
                                         function_lists_file=args.function_list,
                                         annot_names_file=args.annot_names_file, orsum_minterm_size=args.orsumMinTermSize,
                                         selected_adjust_pvalue_cutoff=args.gseapyCutOff, taxon_id_file=args.taxon_id)

    logger.info("--- Total runtime %.2f seconds ---" % (time.time() - start_time))
    logger.warning(f'--- Logs written in {log_file_path} ---')


if __name__ == '__main__':
    main()



import csv
import os
import shutil
import subprocess
import hashlib

from esmecata.core.precomputed import precomputed_parse_affiliation


EXPECTED_RESULTS = {
    'Cluster_1': {'proteomes': 18, 'protein_clusters': 573, 'GOs': 2557, 'ECs': 367}
}

EXPECTED_RESULTS_TWO_DB = {
    'Cluster_1': {'proteomes': 18, 'protein_clusters': 573, 'GOs': 2557, 'ECs': 367},
    'Cluster_2': {'proteomes': 5, 'protein_clusters': 708, 'GOs': 1439, 'ECs': 296}
}

EXPECTED_MD5_BUCHNERA = "501f1308d5954f796155004b1e6bfd44"
EXPECTED_MD5_SECOND = "0a6331f5ea548b1ba6853cfb4c760602"

def test_precomputed_parse_affiliation_offline():
    output_folder = 'output_folder'
    input_file = os.path.join('test_data', 'buchnera_workflow.tsv')
    esmecata_precomputed_db = os.path.join('test_data', 'buchnera_database.zip')
    # Check md5 database.
    md5_database = hashlib.md5(open(esmecata_precomputed_db,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_BUCHNERA

    precomputed_parse_affiliation(input_file, esmecata_precomputed_db, output_folder)

    proteome_tax_id_file = os.path.join(output_folder, '1_clustering', 'proteome_tax_id.tsv')
    clustering_stat_file = os.path.join(output_folder, '1_clustering', 'stat_number_clustering.tsv')
    annotation_stat_file = os.path.join(output_folder, '2_annotation', 'stat_number_annotation.tsv')

    predicted_results = {}
    with open(proteome_tax_id_file, 'r') as open_proteome_tax_id_file:
        csvreader = csv.DictReader(open_proteome_tax_id_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name] = {}
            predicted_results[observation_name]['proteomes'] = len(line['proteome'].split(','))

    with open(clustering_stat_file, 'r') as open_clustering_stat_file:
        csvreader = csv.DictReader(open_clustering_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['protein_clusters'] = int(line['Number_protein_clusters_kept'])

    with open(annotation_stat_file, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['GOs'] = int(line['Number_go_terms'])
            predicted_results[observation_name]['ECs'] = int(line['Number_ecs'])

    for observation_name in EXPECTED_RESULTS:
        for data in EXPECTED_RESULTS[observation_name]:
            assert predicted_results[observation_name][data] == EXPECTED_RESULTS[observation_name][data]

    shutil.rmtree(output_folder)


def test_precomputed_parse_taxon_id_offline():
    output_folder = 'output_folder'
    input_file = os.path.join('test_data', 'buchnera_workflow_ncb_taxid.tsv')
    esmecata_precomputed_db = os.path.join('test_data', 'buchnera_database.zip')
    # Check md5 database.
    md5_database = hashlib.md5(open(esmecata_precomputed_db,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_BUCHNERA

    precomputed_parse_affiliation(input_file, esmecata_precomputed_db, output_folder)

    proteome_tax_id_file = os.path.join(output_folder, '1_clustering', 'proteome_tax_id.tsv')
    clustering_stat_file = os.path.join(output_folder, '1_clustering', 'stat_number_clustering.tsv')
    annotation_stat_file = os.path.join(output_folder, '2_annotation', 'stat_number_annotation.tsv')

    predicted_results = {}
    with open(proteome_tax_id_file, 'r') as open_proteome_tax_id_file:
        csvreader = csv.DictReader(open_proteome_tax_id_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name] = {}
            predicted_results[observation_name]['proteomes'] = len(line['proteome'].split(','))

    with open(clustering_stat_file, 'r') as open_clustering_stat_file:
        csvreader = csv.DictReader(open_clustering_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['protein_clusters'] = int(line['Number_protein_clusters_kept'])

    with open(annotation_stat_file, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['GOs'] = int(line['Number_go_terms'])
            predicted_results[observation_name]['ECs'] = int(line['Number_ecs'])

    for observation_name in EXPECTED_RESULTS:
        for data in EXPECTED_RESULTS[observation_name]:
            assert predicted_results[observation_name][data] == EXPECTED_RESULTS[observation_name][data]

    shutil.rmtree(output_folder)


def test_precomputed_parse_affiliation_cli_offline():
    output_folder = 'output_folder'
    input_file = os.path.join('test_data', 'buchnera_workflow.tsv')
    esmecata_precomputed_db = os.path.join('test_data', 'buchnera_database.zip')
    # Check md5 database.
    md5_database = hashlib.md5(open(esmecata_precomputed_db,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_BUCHNERA

    subprocess.call(['esmecata', 'precomputed', '-i', input_file, '-d', esmecata_precomputed_db, '-o', output_folder])

    proteome_tax_id_file = os.path.join(output_folder, '1_clustering', 'proteome_tax_id.tsv')
    clustering_stat_file = os.path.join(output_folder, '1_clustering', 'stat_number_clustering.tsv')
    annotation_stat_file = os.path.join(output_folder, '2_annotation', 'stat_number_annotation.tsv')

    predicted_results = {}
    with open(proteome_tax_id_file, 'r') as open_proteome_tax_id_file:
        csvreader = csv.DictReader(open_proteome_tax_id_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name] = {}
            predicted_results[observation_name]['proteomes'] = len(line['proteome'].split(','))

    with open(clustering_stat_file, 'r') as open_clustering_stat_file:
        csvreader = csv.DictReader(open_clustering_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['protein_clusters'] = int(line['Number_protein_clusters_kept'])

    with open(annotation_stat_file, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['GOs'] = int(line['Number_go_terms'])
            predicted_results[observation_name]['ECs'] = int(line['Number_ecs'])

    for observation_name in EXPECTED_RESULTS:
        for data in EXPECTED_RESULTS[observation_name]:
            assert predicted_results[observation_name][data] == EXPECTED_RESULTS[observation_name][data]

    shutil.rmtree(output_folder)


def test_precomputed_parse_affiliation_offline_two_databases():
    output_folder = 'output_folder'
    input_file = os.path.join('test_data', 'two_db_precomputed.tsv')
    esmecata_path_buchnera_database = os.path.join('test_data', 'buchnera_database.zip')
    esmecata_path_second_database = os.path.join('test_data', 'second_database.zip')
    esmecata_precomputed_db = esmecata_path_buchnera_database + ' ' + esmecata_path_second_database
    # Check md5 database.
    md5_database = hashlib.md5(open(esmecata_path_buchnera_database,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_BUCHNERA
    md5_database = hashlib.md5(open(esmecata_path_second_database,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_SECOND

    precomputed_parse_affiliation(input_file, esmecata_precomputed_db, output_folder)

    proteome_tax_id_file = os.path.join(output_folder, '1_clustering', 'proteome_tax_id.tsv')
    clustering_stat_file = os.path.join(output_folder, '1_clustering', 'stat_number_clustering.tsv')
    annotation_stat_file = os.path.join(output_folder, '2_annotation', 'stat_number_annotation.tsv')

    predicted_results = {}
    with open(proteome_tax_id_file, 'r') as open_proteome_tax_id_file:
        csvreader = csv.DictReader(open_proteome_tax_id_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name] = {}
            predicted_results[observation_name]['proteomes'] = len(line['proteome'].split(','))

    with open(clustering_stat_file, 'r') as open_clustering_stat_file:
        csvreader = csv.DictReader(open_clustering_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['protein_clusters'] = int(line['Number_protein_clusters_kept'])

    with open(annotation_stat_file, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['GOs'] = int(line['Number_go_terms'])
            predicted_results[observation_name]['ECs'] = int(line['Number_ecs'])

    for observation_name in EXPECTED_RESULTS_TWO_DB:
        for data in EXPECTED_RESULTS_TWO_DB[observation_name]:
            assert predicted_results[observation_name][data] == EXPECTED_RESULTS_TWO_DB[observation_name][data]

    shutil.rmtree(output_folder)


def test_precomputed_parse_affiliation_two_databases_cli_offline():
    output_folder = 'output_folder'
    input_file = os.path.join('test_data', 'two_db_precomputed.tsv')
    esmecata_path_buchnera_database = os.path.join('test_data', 'buchnera_database.zip')
    esmecata_path_second_database = os.path.join('test_data', 'second_database.zip')
    esmecata_precomputed_db = esmecata_path_buchnera_database + ' ' + esmecata_path_second_database
    # Check md5 database.
    md5_database = hashlib.md5(open(esmecata_path_buchnera_database,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_BUCHNERA
    md5_database = hashlib.md5(open(esmecata_path_second_database,'rb').read()).hexdigest()
    assert md5_database == EXPECTED_MD5_SECOND

    subprocess.call(['esmecata', 'precomputed', '-i', input_file, '-d', esmecata_precomputed_db, '-o', output_folder])

    proteome_tax_id_file = os.path.join(output_folder, '1_clustering', 'proteome_tax_id.tsv')
    clustering_stat_file = os.path.join(output_folder, '1_clustering', 'stat_number_clustering.tsv')
    annotation_stat_file = os.path.join(output_folder, '2_annotation', 'stat_number_annotation.tsv')

    predicted_results = {}
    with open(proteome_tax_id_file, 'r') as open_proteome_tax_id_file:
        csvreader = csv.DictReader(open_proteome_tax_id_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name] = {}
            predicted_results[observation_name]['proteomes'] = len(line['proteome'].split(','))

    with open(clustering_stat_file, 'r') as open_clustering_stat_file:
        csvreader = csv.DictReader(open_clustering_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['protein_clusters'] = int(line['Number_protein_clusters_kept'])

    with open(annotation_stat_file, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            predicted_results[observation_name]['GOs'] = int(line['Number_go_terms'])
            predicted_results[observation_name]['ECs'] = int(line['Number_ecs'])

    for observation_name in EXPECTED_RESULTS_TWO_DB:
        for data in EXPECTED_RESULTS_TWO_DB[observation_name]:
            assert predicted_results[observation_name][data] == EXPECTED_RESULTS_TWO_DB[observation_name][data]

    organism_not_found_in_database_file_path = os.path.join(output_folder, 'organism_not_found_in_database.tsv')
    missed_observation_names = []
    with open(organism_not_found_in_database_file_path, 'r') as open_annotation_stat_file:
        csvreader = csv.DictReader(open_annotation_stat_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            missed_observation_names.append(observation_name)

    # All input observation names should be found.
    assert missed_observation_names == []

    shutil.rmtree(output_folder)


if __name__ == "__main__":
    test_precomputed_parse_affiliation_offline()
    test_precomputed_parse_affiliation_cli_offline()
